@@ -2,16 +2,18 @@
 // SPDX-License-Identifier: LGPL-2.1-or-later
 
 #include "todosortfilterproxymodel.h"
+#include "../filter.h"
 
 TodoSortFilterProxyModel::TodoSortFilterProxyModel(QObject *parent)
     : QSortFilterProxyModel(parent)
 {
     const QString todoMimeType = QStringLiteral("application/x-vnd.akonadi.calendar.todo");
-    m_todoTreeModel = new Akonadi::IncidenceTreeModel(QStringList() << todoMimeType, this);
+    m_todoTreeModel.reset(new Akonadi::IncidenceTreeModel(QStringList() << todoMimeType, this));
+
     const auto pref = EventViews::PrefsPtr(new EventViews::Prefs);
-    m_baseTodoModel = new TodoModel(pref, this);
-    m_baseTodoModel->setSourceModel(m_todoTreeModel);
-    setSourceModel(m_baseTodoModel);
+    m_baseTodoModel.reset(new TodoModel(pref, this));
+    m_baseTodoModel->setSourceModel(m_todoTreeModel.data());
+    setSourceModel(m_baseTodoModel.data());
 
     setDynamicSortFilter(true);
     setSortCaseSensitivity(Qt::CaseInsensitive);
@@ -20,16 +22,13 @@ TodoSortFilterProxyModel::TodoSortFilterProxyModel(QObject *parent)
     KSharedConfig::Ptr config = KSharedConfig::openConfig();
     KConfigGroup rColorsConfig(config, "Resources Colors");
     m_colorWatcher = KConfigWatcher::create(config);
-
     QObject::connect(m_colorWatcher.data(), &KConfigWatcher::configChanged, this, &TodoSortFilterProxyModel::loadColors);
 
     loadColors();
-}
 
-TodoSortFilterProxyModel::~TodoSortFilterProxyModel()
-{
-    delete m_baseTodoModel;
-    delete m_todoTreeModel;
+    m_dateRefreshTimer.setInterval(m_dateRefreshTimerInterval);
+    m_dateRefreshTimer.callOnTimeout(this, &TodoSortFilterProxyModel::updateDateLabels);
+    m_dateRefreshTimer.start();
 }
 
 int TodoSortFilterProxyModel::columnCount(const QModelIndex &) const
@@ -43,6 +42,7 @@ QHash<int, QByteArray> TodoSortFilterProxyModel::roleNames() const
     roleNames[TodoModel::SummaryRole] = "text";
     roleNames[Roles::StartTimeRole] = "startTime";
     roleNames[Roles::EndTimeRole] = "endTime";
+    roleNames[Roles::DisplayDueDateRole] = "displayDueDate";
     roleNames[Roles::LocationRole] = "location";
     roleNames[Roles::AllDayRole] = "allDay";
     roleNames[Roles::ColorRole] = "color";
@@ -62,9 +62,9 @@ QHash<int, QByteArray> TodoSortFilterProxyModel::roleNames() const
     roleNames[Roles::CategoriesRole] = "todoCategories"; // Simply 'categories' causes issues
     roleNames[Roles::CategoriesDisplayRole] = "categoriesDisplay";
     roleNames[Roles::TreeDepthRole] = "treeDepth";
-    roleNames[Roles::TopMostParentDueDate] = "topMostParentDueDate";
-    roleNames[Roles::TopMostParentSummary] = "topMostParentSummary";
-    roleNames[Roles::TopMostParentPriority] = "topMostParentPriority";
+    roleNames[Roles::TopMostParentDueDateRole] = "topMostParentDueDate";
+    roleNames[Roles::TopMostParentSummaryRole] = "topMostParentSummary";
+    roleNames[Roles::TopMostParentPriorityRole] = "topMostParentPriority";
 
     return roleNames;
 }
@@ -98,6 +98,8 @@ QVariant TodoSortFilterProxyModel::data(const QModelIndex &index, int role) cons
         return todoPtr->dtStart();
     } else if (role == Roles::EndTimeRole) {
         return todoPtr->dtDue();
+    } else if (role == Roles::DisplayDueDateRole) {
+        return todoDueDateDisplayString(todoPtr, DisplayDateTimeAndIfOverdue);
     } else if (role == Roles::LocationRole) {
         return todoPtr->location();
     } else if (role == Roles::AllDayRole) {
@@ -143,7 +145,7 @@ QVariant TodoSortFilterProxyModel::data(const QModelIndex &index, int role) cons
         return todoPtr->categories();
     } else if (role == Roles::CategoriesDisplayRole) {
         return todoPtr->categories().join(i18nc("List separator", ", "));
-    } else if (role == Roles::TreeDepthRole || role == TopMostParentSummary || role == TopMostParentDueDate || role == TopMostParentPriority) {
+    } else if (role == Roles::TreeDepthRole || role == TopMostParentSummaryRole || role == TopMostParentDueDateRole || role == TopMostParentPriorityRole) {
         int depth = 0;
         auto idx = index;
         while (idx.parent().isValid()) {
@@ -156,32 +158,130 @@ QVariant TodoSortFilterProxyModel::data(const QModelIndex &index, int role) cons
         switch (role) {
         case Roles::TreeDepthRole:
             return depth;
-        case TopMostParentSummary:
+        case TopMostParentSummaryRole:
             return todo->summary();
-        case TopMostParentDueDate: {
+        case TopMostParentDueDateRole: {
             if (!todo->hasDueDate()) {
                 return i18n("No set date");
             }
 
-            const auto isOverdue =
-                (todo->allDay() && todo->dtDue().date() < QDate::currentDate()) || (!todo->allDay() && todo->dtDue() < QDateTime::currentDateTime());
-
-            if (isOverdue) {
+            if (todo->isOverdue()) {
                 return i18n("Overdue");
             }
 
             const auto dateInCurrentTZ = todo->dtDue().toLocalTime().date();
             const auto isToday = dateInCurrentTZ == QDate::currentDate();
 
-            return isToday ? i18n("Today") : todo->dtDue().toString();
+            return isToday ? i18n("Today") : todoDueDateDisplayString(todo, DisplayDateOnly);
         }
-        case TopMostParentPriority:
+        case TopMostParentPriorityRole:
             return todo->priority();
         }
     }
     return QSortFilterProxyModel::data(index, role);
 }
 
+QString TodoSortFilterProxyModel::todoDueDateDisplayString(const KCalendarCore::Todo::Ptr todo,
+                                                           const TodoSortFilterProxyModel::DueDateDisplayFormat format) const
+{
+    if (!todo || !todo->hasDueDate()) {
+        return {};
+    }
+
+    const auto systemLocale = QLocale::system();
+    const auto includeTime = !todo->allDay() && format != DisplayDateOnly;
+    const auto includeOverdue = todo->isOverdue() && format == DisplayDateTimeAndIfOverdue;
+
+    const auto todoDateTimeDue = todo->dtDue().toLocalTime();
+    const auto todoDateDue = todoDateTimeDue.date();
+    const auto todoTimeDueString =
+        includeTime ? i18nc("Please retain space", " at %1", systemLocale.toString(todoDateTimeDue.time(), QLocale::NarrowFormat)) : QStringLiteral(" ");
+    const auto todoOverdueString = includeOverdue ? i18nc("Please retain parenthesis and space", " (overdue)") : QString();
+
+    const auto currentDate = QDate::currentDate();
+    const auto dateFormat = todoDateDue.year() == currentDate.year() ? QStringLiteral("dddd dd MMMM") : QStringLiteral("dddd dd MMMM yyyy");
+
+    static constexpr char translationExplainer[] =
+        "No spaces -- the (optional) %1 string, which includes the time, includes this space"
+        " as does the %2 string which is the overdue string (also optional!)";
+
+    if (currentDate == todoDateDue) {
+        return i18nc(translationExplainer, "Today%1%2", todoTimeDueString, todoOverdueString);
+    } else if (currentDate.daysTo(todoDateDue) == 1) {
+        return i18nc(translationExplainer, "Tomorrow%1%2", todoTimeDueString, todoOverdueString);
+    } else if (currentDate.daysTo(todoDateDue) == -1) {
+        return i18nc(translationExplainer, "Yesterday%1%2", todoTimeDueString, todoOverdueString);
+    }
+
+    const auto dateDueString = systemLocale.toString(todoDateDue, dateFormat);
+    return dateDueString + todoTimeDueString + todoOverdueString;
+}
+
+void TodoSortFilterProxyModel::updateDateLabels()
+{
+    if (rowCount() == 0 || !sourceModel()) {
+        return;
+    }
+
+    emitDateDataChanged({});
+    sortTodoModel();
+    m_lastDateRefreshDate = QDate::currentDate();
+}
+
+void TodoSortFilterProxyModel::emitDateDataChanged(const QModelIndex &idx)
+{
+    const auto idxRowCount = rowCount(idx);
+    const auto srcModel = sourceModel();
+
+    if (idxRowCount == 0) {
+        return;
+    }
+
+    const auto bottomRow = idxRowCount - 1;
+    const auto currentDate = QDate::currentDate();
+    const auto currentDateTime = QDateTime::currentDateTime();
+
+    const auto iterateOverChildren = [this, &idx, &srcModel, &currentDate, &currentDateTime](const int row) {
+        const auto childIdx = index(row, 0, idx);
+
+        const auto todo = childIdx.data(TodoModel::TodoPtrRole).value<KCalendarCore::Todo::Ptr>();
+        const auto isOverdue = todo->isOverdue();
+        const auto dtDue = todo->dtDue();
+        const auto isRecentlyOverdue = isOverdue && currentDateTime.msecsTo(dtDue) >= -m_dateRefreshTimerInterval;
+
+        if (isRecentlyOverdue || m_lastDateRefreshDate != currentDate) {
+            Q_EMIT dataChanged(childIdx, childIdx, {DisplayDueDateRole, TopMostParentDueDateRole});
+
+            // For the proxy model to re-sort items into their correct section we also need to emit a
+            // dataChanged() signal for the column we are sorting by in the source model
+            const auto srcChildIdx = mapToSource(childIdx).siblingAtColumn(TodoModel::DueDateColumn);
+            Q_EMIT srcModel->dataChanged(srcChildIdx, srcChildIdx, {TodoModel::DueDateRole});
+        }
+
+        // We recursively do the same for children
+        emitDateDataChanged(childIdx);
+    };
+
+    // This is a workaround for weird sorting behaviour. If one of the items changes because it becomes
+    // overdue, for example, the way in which we emit dataChanged() signals of the sourceModel will
+    // dictate how the model gets sorted.
+    //
+    // Example: In a case where the sort is ascending (i.e. overdue at top), a 0 to rowCount() -1 sort
+    // will move the overdue item up by only one row due to how the QSortFilterProxyModel uses lessThan().
+    // If we go the opposite way then the QSFPM calls lessThan() on the overdue item more than once, moving
+    // it upwards.
+
+    if (m_sortAscending) {
+        for (auto i = bottomRow; i >= 0; --i) {
+            iterateOverChildren(i);
+        }
+    } else {
+        for (auto i = 0; i < idxRowCount; ++i) {
+            iterateOverChildren(i);
+        }
+    }
+}
+
 bool TodoSortFilterProxyModel::filterAcceptsRow(int row, const QModelIndex &sourceParent) const
 {
     if (filterAcceptsRowCheck(row, sourceParent)) {
@@ -210,15 +310,15 @@ bool TodoSortFilterProxyModel::filterAcceptsRowCheck(int row, const QModelIndex
     const QModelIndex sourceIndex = sourceModel()->index(row, 0, sourceParent);
     Q_ASSERT(sourceIndex.isValid());
 
-    if (m_filterMap.empty()) {
+    if (m_filterObject == nullptr) {
         return QSortFilterProxyModel::filterAcceptsRow(row, sourceParent);
     }
 
     bool acceptRow = true;
 
-    if (m_filterMap.contains(QLatin1String("collectionId")) && m_filterMap[QLatin1String("collectionId")].toInt() > -1) {
+    if (m_filterObject->collectionId() > -1) {
         const auto collectionId = sourceIndex.data(TodoModel::TodoRole).value<Akonadi::Item>().parentCollection().id();
-        acceptRow = acceptRow && collectionId == m_filterMap[QLatin1String("collectionId")].toInt();
+        acceptRow = acceptRow && collectionId == m_filterObject->collectionId();
     }
 
     switch (m_showCompleted) {
@@ -232,8 +332,8 @@ bool TodoSortFilterProxyModel::filterAcceptsRowCheck(int row, const QModelIndex
         break;
     }
 
-    if (m_filterMap.contains(QLatin1String("tags")) && !m_filterMap[QLatin1String("tags")].toStringList().isEmpty()) {
-        const auto tags = m_filterMap[QLatin1String("tags")].toStringList();
+    if (!m_filterObject->tags().isEmpty()) {
+        const auto tags = m_filterObject->tags();
         bool containsTag = false;
         for (const auto &tag : tags) {
             const auto todoPtr = sourceIndex.data(TodoModel::TodoPtrRole).value<KCalendarCore::Todo::Ptr>();
@@ -270,22 +370,21 @@ bool TodoSortFilterProxyModel::hasAcceptedChildren(int row, const QModelIndex &s
     return false;
 }
 
-Akonadi::ETMCalendar::Ptr TodoSortFilterProxyModel::calendar()
+Akonadi::ETMCalendar::Ptr TodoSortFilterProxyModel::calendar() const
 {
     return m_calendar;
 }
 
 void TodoSortFilterProxyModel::setCalendar(Akonadi::ETMCalendar::Ptr &calendar)
 {
-    beginResetModel();
+    // No need to manually emit beginResetModel(), source model does it for us
     m_calendar = calendar;
     m_todoTreeModel->setSourceModel(calendar->model());
     m_baseTodoModel->setCalendar(m_calendar);
     Q_EMIT calendarChanged();
-    endResetModel();
 }
 
-Akonadi::IncidenceChanger *TodoSortFilterProxyModel::incidenceChanger()
+Akonadi::IncidenceChanger *TodoSortFilterProxyModel::incidenceChanger() const
 {
     return m_lastSetChanger;
 }
@@ -298,7 +397,7 @@ void TodoSortFilterProxyModel::setIncidenceChanger(Akonadi::IncidenceChanger *ch
     Q_EMIT incidenceChangerChanged();
 }
 
-void TodoSortFilterProxyModel::setColorCache(QHash<QString, QColor> colorCache)
+void TodoSortFilterProxyModel::setColorCache(const QHash<QString, QColor> colorCache)
 {
     m_colors = colorCache;
 }
@@ -334,24 +433,50 @@ void TodoSortFilterProxyModel::setShowCompleted(int showCompleted)
     sortTodoModel();
 }
 
-QVariantMap TodoSortFilterProxyModel::filterMap() const
+Filter *TodoSortFilterProxyModel::filterObject() const
 {
-    return m_filterMap;
+    return m_filterObject;
 }
 
-void TodoSortFilterProxyModel::setFilterMap(const QVariantMap &filterMap)
+void TodoSortFilterProxyModel::setFilterObject(Filter *filterObject)
 {
-    Q_EMIT filterMapAboutToChange();
-
-    Q_EMIT layoutAboutToBeChanged();
-    m_filterMap = filterMap;
-    Q_EMIT filterMapChanged();
-
-    if (m_filterMap.contains(QLatin1String("name"))) {
-        const auto name = m_filterMap[QLatin1String("name")].toString();
-        setFilterFixedString(name);
+    if (m_filterObject == filterObject) {
+        return;
     }
+
+    if (m_filterObject) {
+        disconnect(m_filterObject, nullptr, this, nullptr);
+    }
+
+    Q_EMIT filterObjectAboutToChange();
+    Q_EMIT layoutAboutToBeChanged();
+    m_filterObject = filterObject;
+    Q_EMIT filterObjectChanged();
+
+    const auto nameFilter = m_filterObject->name();
+    const auto handleFilterNameChange = [this] {
+        Q_EMIT filterObjectAboutToChange();
+        setFilterFixedString(m_filterObject->name());
+        Q_EMIT layoutChanged();
+        Q_EMIT filterObjectChanged();
+    };
+    const auto handleFilterObjectChange = [this] {
+        Q_EMIT filterObjectAboutToChange();
+        invalidateFilter();
+        Q_EMIT layoutChanged();
+        Q_EMIT filterObjectChanged();
+    };
+
+    connect(m_filterObject, &Filter::nameChanged, this, handleFilterNameChange);
+    connect(m_filterObject, &Filter::tagsChanged, this, handleFilterObjectChange);
+    connect(m_filterObject, &Filter::collectionIdChanged, this, handleFilterObjectChange);
+
+    if (!nameFilter.isEmpty()) {
+        setFilterFixedString(nameFilter);
+    }
+
     invalidateFilter();
+
     Q_EMIT layoutChanged();
 
     sortTodoModel();
@@ -363,7 +488,7 @@ void TodoSortFilterProxyModel::sortTodoModel()
     QSortFilterProxyModel::sort(m_sortColumn, order);
 }
 
-void TodoSortFilterProxyModel::filterTodoName(QString name, int showCompleted)
+void TodoSortFilterProxyModel::filterTodoName(const QString &name, const int showCompleted)
 {
     Q_EMIT layoutAboutToBeChanged();
     setFilterFixedString(name);
@@ -464,6 +589,13 @@ int TodoSortFilterProxyModel::compareDueDates(const QModelIndex &left, const QMo
         return 0;
     }
 
+    const auto leftOverdue = leftTodo->isOverdue();
+    const auto rightOverdue = rightTodo->isOverdue();
+
+    if (leftOverdue != rightOverdue) {
+        return leftOverdue ? -1 : 1;
+    }
+
     const bool leftIsEmpty = !leftTodo->hasDueDate();
     const bool rightIsEmpty = !rightTodo->hasDueDate();
 
@@ -473,15 +605,11 @@ int TodoSortFilterProxyModel::compareDueDates(const QModelIndex &left, const QMo
     } else if (!leftIsEmpty) { // Both have due dates
         const auto leftDateTime = leftTodo->dtDue();
         const auto rightDateTime = rightTodo->dtDue();
-        const auto leftOverdue = leftTodo->isOverdue();
-        const auto rightOverdue = rightTodo->isOverdue();
 
-        if (leftDateTime == rightDateTime && leftOverdue == rightOverdue) {
+        if (leftDateTime == rightDateTime) {
             return 0;
-        } else if (leftOverdue == rightOverdue) {
-            return leftDateTime < rightDateTime ? -1 : 1;
         } else {
-            return leftOverdue ? -1 : 1;
+            return leftDateTime < rightDateTime ? -1 : 1;
         }
     } else { // Neither has a due date
         return 0;
